#!/bin/env python3
# the definition of minst class

class MInstClass:
  """Multiple Instance class"""
  ## class members
  instid=""; # instance ID
  chrdata=""; # chromosome
  dirdata=""; # direction
  
  nsegment=0; # number of segments
  nsample=0; # number of samples
  
  segdata=[]; # segment definition: a list of pairs
  segcntdata=[]; # segment count definition: a list of lists
  ntotalrd=[];  # total number of reads; a list
  sgtype=[]; # single-end read types; a list of lists
  sgtypedir=[]; # read type direction; a list
  sgtypecnt=[]; # sg type count. a list of lists
  # intermediate data structures
  connmap=[]; # connectivity map showing the connection between nodes; a list of lists. Connect to itself.
  splicemap=[]; # splice map showing the position of the splicing. do not connect to itself.
  seglen=[]; # length of the segment
  
  ## class functions
  def getConnMap(self): 
    # define a connectivity map
    if len(self.connmap)>0:
      return self.connmap;
    self.connmap=[[0]*self.nsegment for i in range(self.nsegment)];
    self.splicemap=[[0]*self.nsegment for i in range(self.nsegment)];
    sgtypetotalcnt=[sum(x) for x in self.sgtypecnt]; # the total number of counts
    for nst in range(len(self.sgtype)):
      istids=[x for x in range(len(self.sgtype[nst])) if self.sgtype[nst][x] == 1]; # the definition
      if len(istids) == 1:
        dx=istids[0];
        self.connmap[dx][dx]=sgtypetotalcnt[nst];
        # self.splicemap[dx][dx]=sgtypetotalcnt[nst];
      if len(istids) >1: 
        # junction reads, enumerate involved segments
        for itt in range(len(istids)-1):
          dx=istids[itt];
          dy=istids[itt+1];
          # connect dx and dy
          if self.splicemap[dx][dy]<sgtypetotalcnt[nst]:
            self.splicemap[dx][dy]=sgtypetotalcnt[nst];
          if self.connmap[dx][dy]<sgtypetotalcnt[nst]:
            self.connmap[dx][dy]=sgtypetotalcnt[nst];
          # connect all previous nodes connecting to dx
          for pdx in range(dx):
            if self.connmap[pdx][dx]>0:
              min_pdxdxdycnt=self.connmap[pdx][dx];
              if min_pdxdxdycnt>sgtypetotalcnt[nst]:
                min_pdxdxdycnt=sgtypetotalcnt[nst];
              if self.connmap[pdx][dy]<min_pdxdxdycnt:
                self.connmap[pdx][dy]=min_pdxdxdycnt;
          # end for pdx
        # end for itt
      # end if
    # end for nst
    return self.connmap;
  # calculate intermediate datasets
  def calculateIntermediate(self):
    self.getConnMap();
    self.seglen=[self.segdata[i][1]-self.segdata[i][0] for i in range(len(self.segdata))];
  def clear(self):
    self.instid=""; # instance ID
    self.chrdata=""; # chromosome
    self.dirdata=""; # direction
    
    self.nsegment=0; # number of segments
    self.nsample=0; # number of samples
    
    self.segdata=[]; # segment definition: a list of pairs
    self.segcntdata=[]; # segment count definition: a list of lists
    self.ntotalrd=[];  # total number of reads; a list
    self.sgtype=[]; # single-end read types; a list of lists
    self.sgtypedir=[];
    self.sgtypecnt=[]; # sg type count. a list of lists
    
    self.connmap=[];
    self.splicemap=[];
    self.seglen=[];
  def __init__(self):
    self.clear();
  # input and output functions
  # output segment definitions to standard BED file
  def segToBed(self, outfile):
    for it in range(len(self.segdata)):
      sgd=self.segdata[it];
      sgcnt=self.segcntdata[it];
      # output the standard BED file
      sgid='inst'+self.instid+'_seg'+str(it)+'_'+'_'.join([str(x) for x in sgcnt]);
      sgscore=sum([int(x) for x in sgcnt]);
      bedrecfield='\t'.join([self.chrdata,str(sgd[0]-1),str(sgd[1]),sgid,str(sgscore),self.dirdata])
      print(bedrecfield,file=outfile); 
  # output segment normalized counts
  def segNormalizedCntToFile(self,outfile):
    # calculate the sum of the reads in this instance
    sgsumcnt=[0]*(self.nsample);
    for tsgtypecnt in self.sgtypecnt:
      sgsumcnt = [sgsumcnt[i]+tsgtypecnt[i] for i in range(len(tsgtypecnt))];  
    # normalized by the median or mean of the reads in this instance
    # sgsumcnt_median=sorted(sgsumcnt)[len(sgsumcnt)//2];
    sgsumcnt_mean=sum(sgsumcnt)*1.00/len(sgsumcnt);
    # calculate factor according to the total read sum
    sg_factor=[1]*(self.nsample);
    if sgsumcnt_mean > 0:
      for i in range(len(sg_factor)):
        if sgsumcnt[i]>0:
          sg_factor[i]=sgsumcnt_mean/sgsumcnt[i];
    # output segment by segment
    for it in range(len(self.segdata)):
      sgd=self.segdata[it];
      sgcnt=self.segcntdata[it];
      # output the standard BED file record
      sgid='inst'+self.instid+'_seg'+str(it)+'_'+'_'.join([str(x) for x in sgcnt]);
      sgscore=sum([int(x) for x in sgcnt]);
      bedrecfield='\t'.join([self.chrdata,str(sgd[0]-1),str(sgd[1]),sgid,str(sgscore),self.dirdata])
      # define the left and right neighbors
      (itleft,itright)=self.neighborNodes(it);
      nbid=','.join([str(x) for x in itleft])+'#'+','.join([str(x) for x in itright])
      # print('Inst'+self.instid+' seg'+str(it)+':'+);
      # output the normalized read coverage for each segment
      itnb=itleft+itright;
      # reads in neighbor 
      totalnbrt=[ sum([self.segcntdata[i][j] for i in itnb]) for j in range(len(sgcnt))]
      # normalized reads
      normsgcnt=[float(sgcnt[i])*sg_factor[i] for i in range(len(sgcnt))];
      normsgcntstr=[str( int((x*1000))/1000.0) for x in normsgcnt];
      # normalized fractions
      normsgcntfrac=[(lambda x,y: x/(x+y) if (x+y)!=0 else 0)(float(sgcnt[i]),totalnbrt[i]) for i in range(len(sgcnt))];
      normsgcntfracstr=[str( int((x*1000))/1000.0) for x in normsgcntfrac];
      segcntfield='\t'.join([bedrecfield,'\t'.join([str(x) for x in sgcnt]),nbid,'\t'.join([str(x) for x in totalnbrt]),'\t'.join(normsgcntfracstr)]);
      print(segcntfield,file=outfile);
  # pair splicing to bed
  def pairSplicingToBed(self,outfile):
    # get test pairs
    gp=self.getTestPairs();
    for (testpairs, testindex) in gp.items():
      out_i=testpairs[0];
      in_j=testpairs[1];
      # for each group in the test pair, get the corresponding sgtype index. 
      # This includes testsgindex which index the junction reads, 
      # and testexonindex, which includes both junction and exon reads.
      # and testflankingindex, which includes sgtype that includes out_i and in_j
      testsgindex=[[] for testgroup in testindex];
      testexonindex=[[] for testgroup in testindex];
      testflankingindex=[[],[]];
      for testgroup_i in range(len(testindex)): # test the group one by one
        testgroup=testindex[testgroup_i];
        for ti in testgroup: # check all items in this group
          for kl_i in range(len(self.sgtype)):
            kl=self.sgtype[kl_i];
            if kl[out_i]==1 and kl[ti]==1:
              testsgindex[testgroup_i].append(kl_i);
            if kl[in_j]==1 and kl[ti]==1:
              testsgindex[testgroup_i].append(kl_i);
            if kl[ti]==1:
              testexonindex[testgroup_i].append(kl_i);
        # end for ti
        if testgroup==[]:
          for kl_i in range(len(self.sgtype)):
            kl=self.sgtype[kl_i];
            if kl[out_i]==1 and kl[in_j]==1:
              testsgindex[testgroup_i].append(kl_i);
              testexonindex[testgroup_i].append(kl_i);
      # end for testgroup_i
      for kl_i in range(len(self.sgtype)): # check the sgtypes supporting the flanking exons
        kl=self.sgtype[kl_i];
        if kl[out_i]==1:
          testflankingindex[0].append(kl_i);
        if kl[in_j]==1:
          testflankingindex[1].append(kl_i);
      #
      # now, sum up all the reads in the group definition in testsgindex
      testsg_readsum=self.sgReadSum(testsgindex); # a nindex*nsample matrix, 
      testexon_readsum=self.sgReadSum(testexonindex); # same as above
      testflanking_readsum=self.sgReadSum(testflankingindex); # 2*nsample
      testflanking_readsum_sum=matrixSums(testflanking_readsum,"col"); # a vector of size nsample
      # calculate the fraction
      ## testsg_frac=[ ([(testsg_readsum_s[k]+pseudocount)/(testsg_sums[k]+pseudocount*len(testindex)) for k in range(self.nsample)]) for testsg_readsum_s in testsg_readsum]; 
      testsg_frac=matrixFracs(testsg_readsum,"col");
      testexon_frac=matrixFracs(testexon_readsum,"col",providedsum=testflanking_readsum_sum);
      #
      # finally, output to text file
      # IDs
      print('\t'.join([self.chrdata,str(self.segdata[0][0]-1),str(self.segdata[-1][-1]),'inst'+self.instid+'_'+str(testpairs[0])+'_'+str(testpairs[1]),self.dirdata]),end='\t',file=outfile);
      # the exon pairs, 
      print(str(testpairs[0])+','+str(testpairs[1]),end='\t',file=outfile);
      # print the definition of this test
      print(str(len(testindex)),end='\t',file=outfile); # number of branches
      print(';'.join([','.join([str(tj) for tj in ti ]) for ti in testindex]),end='\t',file=outfile) # branch index
      # print the fraction of the branch junction 
      for ti in range(self.nsample):
        print(','.join([str(int(testsg_frac_s[ti]*1000)/1000.0) for testsg_frac_s in testsg_frac ]),end='\t',file=outfile);
      # print the junction read count of each group in each sample
      for ti in range(self.nsample):
        print(','.join([str(testsg_readsum[i][ti]) for i in range(len(testindex))]),end='\t',file=outfile);
      # print the fraction of the exon read count 
      for ti in range(self.nsample):
        print(','.join([str(int(testexon_frac_s[ti]*1000)/1000.0) for testexon_frac_s in testexon_frac ]),end='\t',file=outfile);
      # print the exon read count
      for ti in range(self.nsample):
        print(','.join([str(testexon_readsum[i][ti]) for i in range(len(testindex))]),end='\t',file=outfile);
      # print the flanking read sum
      print('\t'.join([str(x) for x in testflanking_readsum_sum]),file=outfile);
    # end for testpairs
  # end def
  # input method
  def readFromFile(self,infile):
    for line in infile:
      # print(line,end='');
      field=line.strip().split();
      if field[0]=='Instance':
        self.clear();
        self.instid=field[1];
        # reset all variables
        nsegment=0;
        nsegcnt=0;
        ntotalreadln=0;
        nsgtype=0;
        nsgtypecnt=0;
      if field[0]=='segment':
        nsegment=int(field[1]);
        self.nsegment=nsegment;
      elif nsegment>0:
        nsegment=nsegment-1;
        self.segdata += [(int(field[1]),int(field[2]))];
      if field[0]=='segcount':
        nsegcnt=int(field[1]);
      elif nsegcnt>0:
        nsegcnt=nsegcnt-1;
        lcnt=[int(x) for x in field[1:]];
        self.segcntdata += [lcnt];
      if field[0]=='sgtype':
        nsgtype=int(field[1]);
      elif nsgtype>0:
        nsgtype=nsgtype-1;
        lsgt=[int(x) for x in field[1:-1]];
        self.sgtype+= [lsgt];
        self.sgtypedir=int(field[-1]);
      if field[0]=='sgcount':
        nsgtypecnt=int(field[1]);
      elif nsgtypecnt>0:
        nsgtypecnt=nsgtypecnt-1;
        lsgcnt=[int(x) for x in field[1:]];
        self.sgtypecnt+= [lsgcnt];
      if field[0]=='totalreads':
        ntotalreadln=int(field[1]);
      elif ntotalreadln>0:
        ntotalreadln=ntotalreadln-1;
        self.ntotalrd=[int(x) for x in field];
        self.nsample=len(self.ntotalrd);
      if field[0]=='Chr':
        self.chrdata=field[1];
      if field[0]=='Dir':
        self.dirdata=field[1];
        self.calculateIntermediate();
        return 0;
    # end of file
    return -1;
  #
  # define the neighbor nodes for each given node ID
  # return a tuple of left nodes and right nodes
  def neighborNodes(self,nid):
    leftnb=[];
    rightnb=[];
    if nid<0 or nid >= self.nsegment:
      return (leftnb,rightnb);
    # the requirement of these neighbor nodes: must be long enough and has gene expression 
    scnp=self.getConnMap();
    # search left neighbors
    for lit in reversed(range(nid-1)):
      if scnp[lit][nid]<=0:
        continue;
      leftnb+=[lit];
      totalseglen=sum([self.seglen[i] for i in leftnb]);
      if totalseglen>100:
        break;
    for lit in range(nid+2,self.nsegment):
      if scnp[nid][lit]<=0:
        continue;
      rightnb+=[lit];
      totalseglen=sum([self.seglen[i] for i in rightnb]);
      if totalseglen>100:
        break;
    return (leftnb,rightnb);
  # 
  # get the nearest merging points for a list of nodes
  # if there are no merging points, return -1
  def getNearestAncestor(self,lpoints):
    if len(lpoints)==0:
      return -1;
    minlp=min(lpoints);
    for lpi in range(minlp,-1,-1):
      noconnect=False;
      for x in lpoints:
        if self.connmap[lpi][x]==0:
          noconnect=True;
          break;
      if noconnect==False:
        return lpi;
    return -1;
  def getNearestDescendant(self,lpoints):
    if len(lpoints)==0:
      return -1;
    minlp=max(lpoints);
    for lpi in range(minlp,self.nsegment):
      noconnect=False;
      for x in lpoints:
        if self.connmap[x][lpi]==0:
          noconnect=True;
          break;
      if noconnect==False:
        return lpi;
    return -1;
  # get test pairs of nodes for alternative splicing
  # only nodes with multiple AS events are returned
  def getTestPairs(self):
    # get branch points: those with >1 outgoing or incoming edges
    ns=self.nsegment;
    outgoingd=[0]*ns;
    incomingd=[0]*ns;
    for it in range(ns):
      for jt in range(ns):
        if jt == it or jt == it-1 or jt == it+1:
          # skip neighbor nodes
          continue;
        if jt< it and self.splicemap[jt][it] >0:
          incomingd[it] += 1;
        if jt > it and self.splicemap[it][jt] >0:
          outgoingd[it] += 1;
    outgoinglist=[x for x in range(ns) if outgoingd[x]>1];
    incominglist=[x for x in range(ns) if incomingd[x]>1];
    # search for each incoming and outgoing edges
    goodlist={};
    for out_i in outgoinglist:
      for in_j in incominglist:
        if in_j <= out_i+1:
          continue;
        # get all sons of out_i which connect to in_j
        out_i_sons=[x for x in range(out_i+2,in_j+1) if self.splicemap[out_i][x]>0 and self.connmap[x][in_j]>0];
        # get all fathers of in_j which connect to out_i
        in_j_fathers=[x for x in range(out_i,in_j-1) if self.splicemap[x][in_j]>0 and self.connmap[out_i][x]>0];
        # check if the nearest ancestor of in_j_fathers is out_i and vise versa
        #print('Testing '+str(out_i)+' '+str(in_j));
        #print('sons:'+' '.join([str(x) for x in out_i_sons]));
        #print('fathers:'+' '.join([str(x) for x in in_j_fathers]));
        #print('ancestor: '+str(self.getNearestAncestor(in_j_fathers))+ ',descendant:'+str(self.getNearestDescendant(out_i_sons)));
        if self.getNearestAncestor(in_j_fathers) != out_i or self.getNearestDescendant(out_i_sons) != in_j:
          continue;
        # now, group the ancestors and descendants into subgroup
        tarlc=out_i_sons[:];
        tarlc.extend(in_j_fathers);
        targetlistcandidate=list(set([x for x in tarlc if x !=out_i and x !=in_j]));
        targetlist=[];
        for x in targetlistcandidate:
          connectedx=False;
          for t in targetlist:
            for ti in t:
              if self.connmap[x][ti]>0 or self.connmap[ti][x]>0:
                connectedx=True;
                break;
              # endif
            # end for ti
            if connectedx==True:
              t.append(x);
              break;
          # end for t
          if connectedx==False:
            targetlist.append([x]);
        # end for x
        if len(targetlist)==0:
          continue;
        if self.splicemap[out_i][in_j]>0:
          targetlist.append([]);
        goodlist[(out_i,in_j)]=targetlist;
      # end for in_j
    # end for out_i
    return goodlist;
  # return the sum of reads in sgtype
  # parameter:
  # sglist: a list of list, each sublist showing the index of the sgtype to be summed up
  # return value: a list of list summing up the sgtype read counts. the size of a list equals to the size of the sglist; and the size of a sublist equals to self.nsample
  def sgReadSum(self,sglist):
    test_readsum=[[0]*self.nsample for sublist in sglist]; # a nindex*nsample matrix, 
    for sglist_i in range(len(sglist)):
      sublist=sglist[sglist_i];
      for sgtype_i in sublist:
        if sgtype_i<0 or sgtype_i >= len(self.sgtypecnt):
          raise Exception('Error: sgtype out of bound.');
        test_readsum[sglist_i]=[test_readsum[sglist_i][k]+self.sgtypecnt[sgtype_i][k] for k in range(self.nsample)];
    return test_readsum;


# for a given matrix (list of lists), calculate the sum
# dir may be "row" or "col". 
# For "row": just summing up the sublists
# For "col": summing up every position in the sublist. Requires the sublist to have the same length
def matrixSums(matlist,dir):
  if dir=="row":
    return [sum(x) for x in matlist];
  elif dir=="col":
    if len(matlist)==0:
      return 0;
    n=len(matlist[0]);
    sgsums=[0]*n;
    for sublist in matlist:
      if len(sublist) != n:
        raise Exception('Error: the length of each sublist is not the same.');
        return sgsums;
    for sublist in matlist:
      sgsums=[sgsums[k]+sublist[k] for k in range(n)];
    return sgsums;
  else:
    raise Exception("Error: unrecognized directions:"+str(dir)+". Must be either row or col.");
    return 0;

# for a given matrix (list of lists), calculate the sum
# dir may be "row" or "col". 
# For "row": just summing up the sublists
# For "col": summing up every position in the sublist. Requires the sublist to have the same length
# if providesum is provided, will use the provided sum to calculate the fraction;
# otherwise, use the sum of rows (or lists)
def matrixFracs(matlist,dir,providedsum=None,pseudocount=5):
  if providedsum is None:
    sumval=matrixSums(matlist,dir);
  else:
    sumval=providedsum;
  if pseudocount ==0 :
    sumval=[ (lambda x: x if x!= 0 else 1)(t) for t in sumval];
  m=len(matlist);
  if dir=="row":
    if len(sumval) != m:
      raise Exception('Error: the sum vectors do not match');
    return [[ (matlist[j][i]+pseudocount)/(sumval[j]+pseudocount*len(matlist[j])) for i in range(len(matlist[j])) ] for j in range(m)];
  elif dir=="col":
    if len(matlist)==0:
      return 0;
    n=len(matlist[0]);
    if len(sumval)!=n:
      raise Exception('Error: incorrect dimension.');
    if len(sumval) != n:
      raise Exception('Error: the sum vectors do not match');
    return [ [(matlist[j][i]+pseudocount)/(sumval[i]+pseudocount*m) for i in range(n) ] for j in range(m)];
  else:
    raise Exception("Error: unrecognized directions:"+str(dir)+". Must be either row or col.");
    return 0;









