/* graph algorithms */
#ifndef GRAPHALGO_H
#define GRAPHALGO_H

#include "commontype.h"
#include "minst.h"
#include "menumiso.h"

/* get the connectivity graph based on the pre-selected parents */
void getcongraph(MInst& inst, mi& allparent, mi& allconn);

/* For each instance, return pairs of segment branches that are subject to correlate. 
  The pair of segment branches are stored in pcheck, each vector contains two numbers, a and b.
  You need to check between all parent chains of node a and all children chains of node b.
  Use getchildrenchain() and getparentchain() to get corresponding segment numbers.
*/
int getcheckingsegments(MInst& inst, mi& rcount,mi& allchildren, mi& allparent, EnumParameter para, mi & pchecks);



/* For each node ns with more than one incoming edges, get the corresponding parents for each edge */
int getparentchain(MInst& inst, mi& allchildren, mi& allparent, int ns, mi& res,EnumParameter para,mi& rcount);
/* For each node ns with more than one outgoing edges, get the corresponding children for each edge */
int getchildrenchain(MInst& inst, mi& allchildren, mi& allparent, int ns, mi& res, EnumParameter para,mi& rcount);


/* Get reference segments: segments that appear in one isoform. */
int getreferencesegments(MInst& inst, mi& rcount, 
  mi& allchildren,
  mi& allparent, 
  EnumParameter para, 
  mi & refs);

/* Get the adjusted children list: remove possible intron childrens. */
vi adjustchildren(mi& allchildren, int tk,MInst& inst, EnumParameter para, mi& rcount);

/* Get the adjusted parent list: remove possible intron parents. */
vi adjustparent(mi& allparent, int tk,MInst& inst, EnumParameter para,mi& rcount);



/* Check if two segments, a and b, are coexpressed in annotation (inst.refs)
Return value: -1 if error, 0 if not, 1 if true. */
int iscoexpressioninanno(MInst& inst, int a, int b);


/* Check if two segments, a and b, are exclusive in annotation (inst.refs)
Return value: -1 if error, 0 if not, 1 if true. */
int isexclusiveinanno(MInst& inst, int a, int b);


/* Get the mean value of a set of vectors */
double getmean(vd& vals);

/* get the std value of a set of vectors */
double getstd(vd& vals);


#endif
