#include <iostream>
#include <fstream>
#include <string>
#include "instance.h"

using namespace std;

string ofile;
vector<string> instfiles;

void printmsg(){
  cout<<"Usage: mergeinstances {OPTIONS} [Instance 1] [Instance 2] ...\n";
  cout<<"Options:\n";
  cout<<"\t-o [filename]\t\tSpecify the output name.\n";
}

int parseopt(int argc,char*argv[]){
  int ii=1;
  while(ii<argc){
    if(string(argv[ii])=="-h"){
      printmsg();
      return -1;
    }else if(string(argv[ii])=="-o" && ii<argc-1){
      //mis.readEJ(argv[ii+1]);
      ofile=string(argv[ii+1]);
      ii++;
    }else{
      instfiles.push_back(string(argv[ii]));
    }
    ii++;
  }
  return 0;
}

int main(int argc, char* argv[]){
  ofile="allpred.ins";
  if(parseopt(argc,argv)==-1)return 0;
  // get the total number of reads
  vl totalreads(instfiles.size()); 
  for(int i=0;i<instfiles.size();i++){
    cout<<"Instance file:"<<instfiles[i]<<endl;
    long tr=gettotalreads(instfiles[i]);
    if(tr==-1){
      return -1;
    }
    totalreads[i]=tr;
    cout<<"Total reads:"<<tr<<endl;
  }
  //open instances
  vector<ifstream*> vif(instfiles.size());
  for(int i=0;i<vif.size();i++){
    vif[i]=new ifstream();
    vif[i]->open(instfiles[i].c_str());
    if(!vif[i]->is_open()){
      cerr<<"Error opening file "<<instfiles[i]<<endl;
      return -1;
    }
  }
  ofstream ofs(ofile.c_str());
  if(!ofs.is_open()){
    cerr<<"Error opening output file "<<ofile<<endl;
  }
  while(true){
    vector<Instance> allv(instfiles.size());
    bool haserror=false;
    for(int i=0;i<vif.size();i++){
      if(allv[i].readFromStream(*vif[i])==-1){
        haserror=true;
        break;
      }
      if(i>0 && allv[0].checkBoundaryWithOtherInst(allv[i])==false){
        cerr<<"Error: instance boundaries of 0 and "<<i<<" are not identical for instance "<<allv[i].id<<".\n";
        haserror=true;
        break;
      }
    }
    if(haserror){
      cerr<<"An error occures.\n";
      break;
    }
    
    if(inst2minst(ofs,allv,totalreads)==-1){
      cerr<<"An error occurs when merging instances.\n";
      break;
    }
    cout<<"Merging instance "<<allv[0].id<<" ...\n";
  }
  ofs.close();
  //close instances
  for(int i=0;i<vif.size();i++) {
    vif[i]->close();
    delete vif[i];
  }

  return 0;
}
