/*
Define multiple instance
*/

#ifndef MINST_H
#define MINST_H

#include <vector>
#include <string>
#include <iostream>
#include "commontype.h" //get the range_t definition
using namespace std;

class MInst{
public:
  /* Segment */
  ml segment;
  mi segcount;
  /* coverage */
  md segcvg;
  md segmaxcvg;
  md segp0;
  /* type, type count and type direction */
  mi type;
  mi typecount;
  vi typedir;

  /* paired-end type, type count, and distance information */
  mi petype;
  mi petypecount;
  mi petypedist;

  /* chromosome name and id */
  string chrname;
  string id;

  /* Direction */
  int dir;

  /* read length */
  vi readlen;
  
  /* Total number of reads */
  vi totalreads;

  /* Reference annotation, if possible */
  mi refs;

  /* clear the content of all containers*/
  void clear();

  int check();

  /* read from file */
  int readFromFile(char* prefix);
  /* Read from stream. 
     Jump instances in two cases:
       1, if oid!="", read the next instance; otherwise, search for instance with id equal to oid.
       2, if db>0, get the next instance with instanceid%dtotal==da. da must be between [0,db-1].
  */
  int readFromStream( istream& ifs,string oid,int da, int dtotal);

  /* convert type to index matrix. 
  The original matrix is 0-1 matrix. After conversion, only index of 1 elements are saved. */
  void converttype();

  /* Get the number of segments and the number of samples */
  int getNSegs(){return segment.size();}
  int getNSamples(){if(segcount.size()==0)return -1; return segcount[0].size();}
  int getReadLen(int n){if(readlen.size()==0 || n>=readlen.size())return -1; return readlen[n];}
  int getSegLen(int n){if(n>=segment.size())return 0; return segment[n][1]-segment[n][0]+1;}

  /* get the number of reads of a given sample */
  int getSampleRdCnt(int n);
  
  /* get the range of the current instance. Return range(-1,-1) if failed */
  range_t getRange();
  /*
  Fill the rcount matrix with direction sensitive
  Parameter: 
    dir=1: positive; -1: negative; 0: don't care
    ns: the specified sample used. If -1, use the average of all samples
  */
  int fillRCount(mi& rcount,int dir,int ns=-1);


  /* External junction direction information 
  */
  static map<string, map<long,int> >ejdir;
  /* Read external junction information from file */
  static int readEJ(char* filename);


};

#endif
